using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Xml;
using System.Net;

using ESRI.ArcGISExplorer;
using ESRI.ArcGISExplorer.Application;
using ESRI.ArcGISExplorer.Mapping;
using ESRI.ArcGISExplorer.Geometry;
using ESRI.ArcGISExplorer.Data;
using ESRI.ArcGISExplorer.Threading;

using GlobeSpotterAPI;

namespace BAPI_ArcGis_RefImpl
{
    public partial class RefImplBAPI : ESRI.ArcGISExplorer.Application.DockWindow, IBAPIClient
    {
        public static String _wfsServer = @"http:\\EXAMPLE"; // Please change this to a working adress (see 'md_ViewpointChanged').
        public static String _keyValue = "key_value_example"; // Please change this to a valid key (see 'OnComponentReady').

        private BAPI _bapi;
        private int _gmlID = -1;
        private Queue<uint> _drawings;
        private int _drawingCounter;
        private Random _rand = new Random();
        private uint _heightEntityID = uint.MaxValue;
        private uint _areaEntityID = uint.MaxValue;
        private Permissions _permissions;

        private ViewerParameters _viewpars = new ViewerParameters(0.0, 0.0, 90.0);

        private MapDisplay _md;
        private List<MapItem> _mdReclocs = new List<MapItem>();
        private static readonly String _MD_TAG = "_cyclo_";
        private static readonly Symbol _MD_SYMBOL = Symbol.Marker.Sphere.Red;

        private PasswordBox _pb = new PasswordBox();
        private DialogResult _dr = DialogResult.Cancel;

        public RefImplBAPI()
        {
            _bapi = new BAPI(InitType.REMOTE);

            InitializeComponent();
            this.pnlView.Controls.Add(_bapi.gui);

            this.listBox1.Items.Add("Select on map");
            this.listBox1.Items.Add("Image ID");
            this.listBox1.Items.Add("Address");
            this.listBox1.Items.Add("Coordinate");
            this.listBox1.SelectedIndex = 0;

            this.textBox1.Text = "5b04llfu";

            this.listBox2.Items.Add("Point");
            this.listBox2.Items.Add("Image");
            this.listBox2.Items.Add("Cross hair");
            this.listBox2.SelectedIndex = 0;

            _drawings = new Queue<uint>();
            _drawingCounter = 1; 
            
            _md = ESRI.ArcGISExplorer.Application.Application.ActiveMapDisplay;
            _MD_SYMBOL.Size = _MD_SYMBOL.Size * 0.667;
            _md.ViewpointChanged += new EventHandler(this.md_ViewpointChanged);
            _md.MapItemClicked += new EventHandler<MapItemMouseEventArgs>(this.md_MapItemClicked);
        }

        public override void OnShow()
        {
            base.OnShow();
            _dr = _pb.ShowDialog();
            _bapi.Initialize(this);
        }

        private void OpenAndZoomToImage(String input, OpenImageType type)
        {
            _bapi.OpenImage(input, type, _viewpars);
        }

        private bool CheckReady()
        {
            bool result = _bapi.GetAPIReadyState();
            if (!result) {
                MessageBox.Show("GlobeSpotter not ready to open image.", "Warning from GlobeSpotter API", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            }
            return result;
        }

        #region IBAPIClient Members

        public void OnComponentReady()
        {
            lblViewer.Text = _bapi.GetApplicationName() + " v" + _bapi.GetMajorVersion() + @"/" + _bapi.GetMinorVersion();

            if(_dr == DialogResult.OK) {
                _keyValue = _pb.APIKey;
                _wfsServer = _pb.WFSServer;
                _bapi.SetUserNamePassword(_pb.UserName, _pb.Password);
            }

            _bapi.SetApplicationParameter(_keyValue, ApplicationParameterType.API_KEY);
            _bapi.SetApplicationParameter("EPSG:28992", ApplicationParameterType.SRS_NAME_MAP);
        }

        public void OnAPIReady()
        {
            _permissions = _bapi.GetPermissions();

            if (_permissions.Basic) {
                pnlView.Enabled = true;
                pnlUI.Enabled = true;

                _bapi.SetViewerClickMode(ViewerClickMode.X_Y);

                button6.Enabled = _permissions.MeasureHeight;
                btnViewshot.Enabled = _permissions.MeasureArea;
                button8.Enabled = _permissions.MeasureVolume;
            } else {
                MessageBox.Show("Authentication for GlobeSpotter failed!", "Warning in GlobeSpotter API", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            }
        }

        public void OnAPINotReady()
        {
            MessageBox.Show("API not ready...");
        }

        public void OnOpenImageResultEmpty(string input, OpenImageType type, ViewerParameters viewDir, uint viewerID)
        {
            MessageBox.Show("Unable to determine an image id based on using an openImage input value of: " + input, "Exception in GlobeSpotter API", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        public void OnOpenImageError(string input, OpenImageType type, ViewerParameters viewDir, uint viewerID)
        {
            MessageBox.Show("An error has occured while attempting to open an image using an openImage input value of: " + input, "Exception in GlobeSpotter API", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        public void OnImageChanged(uint viewerID)
        {
            RecordingLocation recloc = _bapi.GetRecordingLocation(viewerID); // viewerID always 0 in Basic API (as opposed to Multi & Multi+Map)

            if (recloc != null) {
                lblViewer.Text = recloc.Id;

                // Compensate for esri/RDNew issue (-20, -100 are simplified & estimated values).
                ESRI.ArcGISExplorer.Geometry.Point point = new ESRI.ArcGISExplorer.Geometry.Point(recloc.X - 20, recloc.Y - 100, CoordinateSystem.ProjectedCoordinateSystems.NationalGrids.Europe.RDNew);
                _md.ZoomTo(point, ZoomLevel.Houses);
            }
        }

        public void OnImagePreviewCompleted(uint viewerID)
        {
        }

        public void OnImageSegmentLoaded(uint viewerID)
        {
        }

        public void OnImageCompleted(uint viewerID)
        {
        }

        public void OnImageFailed(uint viewerID)
        {
        }

        public void OnViewLoaded(uint viewerID)
        {
        }

        public void OnViewChanged(uint viewerID, double yaw, double pitch, double hFov)
        {
        }

        public void OnViewerClicked(uint viewerID, double[] mouseCoords)
        {
            if (layerCheckBox.Checked) {
                DrawingMode mode = (DrawingMode)(listBox2.SelectedIndex + 1);
                _bapi.SetDrawingMode(mode);

                if (mode == DrawingMode.URL_IMAGE) {
                    _bapi.SetDrawingImageURL(@"http://www.cyclomedia.nl/images/fisheye1.gif");
                }

                _bapi.SetDrawingSize(_rand.NextDouble() * 16.0 + 8.0);
                Color color = Color.FromArgb((int)(254.0 * _rand.NextDouble()), (int)(254.0 * _rand.NextDouble()), (int)(254.0 * _rand.NextDouble()));
                _bapi.SetDrawingColor(color);
                _drawingCounter++;

                uint drawingID = 0;

                try {
                    switch (_bapi.GetViewerClickMode()) {
                        case ViewerClickMode.X_Y: {
                                drawingID = _bapi.DrawMarkerAtXY(0, mouseCoords[0], mouseCoords[1], "" + _drawingCounter);
                            } break;

                        case ViewerClickMode.H_V: {
                                drawingID = _bapi.DrawMarkerAtHV(0, mouseCoords[0], mouseCoords[1], "" + _drawingCounter);
                            } break;

                        case ViewerClickMode.VECTOR: {
                                drawingID = _bapi.DrawMarkerInDirection(0, mouseCoords[0], mouseCoords[1], mouseCoords[2], "" + _drawingCounter);
                            } break;
                    }
                } catch (Exception ex) {
                    Console.WriteLine(ex);
                }

                _drawings.Enqueue(drawingID);
                if (_drawings.Count > 10) {
                    uint id = _drawings.Dequeue();
                    _bapi.ClearDrawing(0, id);
                }

            }
        }

        public void OnMarkerClicked(uint viewerID, uint drawingID, double[] markerCoords)
        {
            _bapi.ClearDrawings(0);
            _drawings.Clear();
        }

        public void OnEntityDataChanged(uint entityID, EntityData data)
        {
        }

        public void OnEntityFocusChanged(int entityID, EntityData data)
        {
        }

        #endregion

        #region GUI Controls

        private void btnSelect_Click(object sender, EventArgs e)
        {
            CheckReady();
            try {
                int imageOpenValue = listBox1.SelectedIndex;
                
                if (imageOpenValue == 0) {
                    MapDisplay md = ESRI.ArcGISExplorer.Application.Application.ActiveMapDisplay;
                    ESRI.ArcGISExplorer.Geometry.Point trackedPoint = md.TrackPoint();
                    ESRI.ArcGISExplorer.Geometry.Point projectedPoint = (ESRI.ArcGISExplorer.Geometry.Point)GeometryOperations.Project(trackedPoint, CoordinateSystem.ProjectedCoordinateSystems.NationalGrids.Europe.RDNew);

                    String request = "(" + projectedPoint.X + "; " + projectedPoint.Y + ")";
                    OpenAndZoomToImage(request, OpenImageType.COORDINATE); //_bapi.OpenImage(request, OpenImageType.COORDINATE);
                    lblViewer.Text = request;
                } else {
                    OpenAndZoomToImage(textBox1.Text, (OpenImageType)imageOpenValue); //_bapi.OpenImage(textBox1.Text, (OpenImageType)imageOpenValue, viewerParams);
                    lblViewer.Text = textBox1.Text;
                }
            } catch (GlobeSpotterAPIException gsapiex) {
                MessageBox.Show("Exception on attempt to open image: " + gsapiex.Message, "Exception in GlobeSpotter API", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void btnViewshot_Click(object sender, EventArgs e)
        {
            CheckReady();
            try {
                Bitmap bmp = _bapi.GetViewshot(0);

                try {
                    SaveFileDialog sfd = new SaveFileDialog();
                    sfd.Title = "Save Viewshot as BitMap";
                    sfd.Filter = "BitMap (*.bmp)|*.bmp";
                    DialogResult dr = sfd.ShowDialog();
                    if (dr == DialogResult.OK || dr == DialogResult.Yes) {
                        bmp.Save(sfd.FileName);
                    }
                    
                } catch (Exception ex) {
                    MessageBox.Show("Exception on save: " + ex.Message, "Exception in GlobeSpotter Example", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }

            } catch (GlobeSpotterAPIException gsapiex) {
                MessageBox.Show("Exception on retrieve: " + gsapiex.Message, "Exception in GlobeSpotter API", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void button2_Click_1(object sender, EventArgs e)
        {
            _bapi.RotateUp(0, 5);
        }

        private void button5_Click_1(object sender, EventArgs e)
        {
            _bapi.RotateLeft(0, 5);
        }

        private void button4_Click_1(object sender, EventArgs e)
        {
            _bapi.RotateRight(0, 5);
        }

        private void button3_Click_1(object sender, EventArgs e)
        {
            _bapi.RotateDown(0, 5);
        }

        private void gammaBar_Scroll(object sender, EventArgs e)
        {
            _bapi.SetGamma(0, 0.01 * (double)gammaBar.Value);
        }

        private void drawingCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            try {
                if (!layerCheckBox.Checked) {
                    _bapi.HideRecordingLocations();
                    _bapi.HideEntityLayer();
                    _bapi.HideDrawingLayer();
                    if (_gmlID >= 0) {
                        _bapi.RemoveGMLLayer((uint)_gmlID);
                    }
                } else {
                    _bapi.ShowRecordingLocations("reclocs"); // Name is required, but can be anything.
                    _bapi.ShowEntityLayer("entities"); // ... same here.
                    _bapi.ShowDrawingLayer("drawings"); // ... and here.
                }
            } catch (Exception ex) {
                MessageBox.Show(ex.Message, "Exception in GlobeSpotter API", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void button11_Click(object sender, EventArgs e)
        {
            try {
                Permissions prms = _bapi.GetPermissions();
                RecordingLocation recloc = _bapi.GetRecordingLocation(0);
                ViewerClickMode vcl = _bapi.GetViewerClickMode();
                double yaw = _bapi.GetYaw(0);
                double pitch = _bapi.GetPitch(0);
                double hFov = _bapi.GetHFov(0);

                MessageBox.Show
                    ((" Permissions: basic=" + prms.Basic + " wfs=" + prms.AddLayerWFS + "\n\n"
                      + " Click-Mode: " + vcl.ToString() + "\n\n"
                      + " Recording Location:\n date=" + recloc.RecordedAt + "\n id=" + recloc.Id + "\n x=" + recloc.X + "(+-" + recloc.SX + ")\n y=" + recloc.Y + "(+-" + recloc.SZ + ")\n z=" + recloc.Z + "(+-" + recloc.SZ + ")\n\n"
                      + " Viewdir.: yaw=" + yaw + " pitch=" + pitch + " hFov=" + hFov
                      )
                    , "GlobeSpotter Recording Location Info", MessageBoxButtons.OK, MessageBoxIcon.Information
                    );
            } catch (Exception ex) {
                MessageBox.Show(ex.Message, "Exception in GlobeSpotter API", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void button9_Click(object sender, EventArgs e)
        {
            _bapi.RemoveAllEntities();
            _areaEntityID = uint.MaxValue;
            _heightEntityID = uint.MaxValue;
        }

        private void button6_Click(object sender, EventArgs e)
        {
            if (_heightEntityID == uint.MaxValue) {
                RecordingLocation rl = _bapi.GetRecordingLocation(0);

                _heightEntityID = _bapi.AddHeightEntity("heightE", 2.2, new Point3D(rl.X + 1, rl.Y - 3, 1.5));
                EntityData dat = _bapi.GetEntityData(_heightEntityID);
            } else {
                _bapi.RemoveEntity(_heightEntityID);
                _heightEntityID = uint.MaxValue;
            }
        }

        private void button8_Click(object sender, EventArgs e)
        {
            if (_areaEntityID == uint.MaxValue) {
                RecordingLocation rl = _bapi.GetRecordingLocation(0);

                _areaEntityID = _bapi.AddAreaEntity("areaE", 2.0, new Point2D(rl.X - 3, rl.Y - 3), new Point2D(rl.X + 3, rl.Y + 3), 0.0);
                EntityData dat = _bapi.GetEntityData(_areaEntityID);
            } else {
                _bapi.RemoveEntity(_areaEntityID);
                _areaEntityID = uint.MaxValue;
            }
        }

        #endregion

        #region Recording Locations on Map

        public void md_ViewpointChanged(object sender, EventArgs e)
        {
            if (!_bapi.GetAPIReadyState()) return;

            foreach (Note note in _mdReclocs) {
                note.RemoveFromParent();
            }

            Envelope bounds = _md.Extent;
            double manhattanRadius = Math.Max(bounds.Width, bounds.Height) / 2;
            if (manhattanRadius > 250) return;

            ESRI.ArcGISExplorer.Geometry.Point center = (ESRI.ArcGISExplorer.Geometry.Point)GeometryOperations.Project(_md.Center, CoordinateSystem.ProjectedCoordinateSystems.NationalGrids.Europe.RDNew);

            String requestTemplateP0 = _wfsServer + @"/wfsservice/wfsservice.svc/wfs?SERVICE=WFS&VERSION=1.1.0&REQUEST=GetFeature&SRSNAME=EPSG:4326&TYPENAME=atlas:Recording&FILTER=<Filter><And><BBOX><gml:Envelope srsName='EPSG:28992'><gml:lowerCorner>";
            String requestTemplateP1 = @"</gml:lowerCorner><gml:upperCorner>";
            String requestTemplateP2 = @"</gml:upperCorner></gml:Envelope></BBOX><PropertyIsNull><PropertyName>expiredAt</PropertyName></PropertyIsNull></And></Filter>&";

            String requestData0 = "" + (center.X - 20 - manhattanRadius) + " " + (center.Y + 100 - manhattanRadius);
            String requestData1 = "" + (center.X - 20 + manhattanRadius) + " " + (center.Y + 100 + manhattanRadius);

            String request = requestTemplateP0 + requestData0 + requestTemplateP1 + requestData1 + requestTemplateP2;

            try {
                WebRequest getRequest = WebRequest.Create(request);
                getRequest.Method = "GET";
                getRequest.Credentials = CredentialCache.DefaultCredentials;

                HttpWebResponse response = (HttpWebResponse)getRequest.GetResponse();
                Stream dataStream = response.GetResponseStream();
                StreamReader reader = new StreamReader(dataStream);
                XmlDocument gml = new XmlDocument();
                gml.Load(reader);

                ESRI.ArcGISExplorer.Geometry.Point pt = null;
                XmlNode fGML = NavigateXML(gml, new String[] { "wfs:FeatureCollection", "gml:featureMembers" });
                if (fGML != null) {
                    foreach (XmlNode rlGML in fGML.ChildNodes) {
                        XmlNode posGML = NavigateXML(rlGML, new String[] { "atlas:location", "gml:Point", "gml:pos" });
                        if (posGML != null) {
                            String[] locsText = posGML.InnerText.Split(' ');
                            pt = new ESRI.ArcGISExplorer.Geometry.Point(double.Parse(locsText[0]), double.Parse(locsText[1]), new CoordinateSystem(4326)); //CoordinateSystem.ProjectedCoordinateSystems.NationalGrids.Europe.RDNew);

                            XmlNode idGML = NavigateXML(rlGML, new String[] { "atlas:imageId" });
                            Note pin = new ESRI.ArcGISExplorer.Mapping.Note(idGML != null ? idGML.InnerText : "unknown", pt, _MD_SYMBOL);
                            pin.UsePopup = false;
                            pin.Tag = _MD_TAG;
                            _md.Map.ChildItems.Add(pin);
                            _mdReclocs.Add(pin);
                        }
                    }
                }

                reader.Close();
                dataStream.Close();
                response.Close();
            } catch (Exception ex) {
                MessageBox.Show("Could not retrieve (some?) Recording Locations on the Map, because " + ex.Message, "Exception in GlobeSpotter", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            }
        }

        public void md_MapItemClicked(object sender, MapItemMouseEventArgs e)
        {
            if (!_bapi.GetAPIReadyState()) return; 

            MapItem mi = e.MapItems[0];
            if (mi != null && mi.Tag.Equals(_MD_TAG)) {
                OpenAndZoomToImage(mi.Name, OpenImageType.IMAGE_ID); //_bapi.OpenImage(mi.Name, OpenImageType.IMAGE_ID);
            }
        }

        #endregion

        private static XmlNode NavigateXML(XmlNode node, String[] path)
        {
            XmlNode result = node;

            for (int i = 0; i < path.Length; ++i) {
                if (result[path[i]] != null) {
                    result = result[path[i]];
                } else {
                    return null;
                }
            }

            return result;
        }
    }
        
}
